/*
 *  player.h
 *  Copyright (C) 1998,1999 SAKAI Katsuya
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifndef __PLAYER_H
#define __PLAYER_H

#include <sys/types.h>
#include <unistd.h>
#include "queue.h"
#include "recomposer.h"
#include "smf.h"
#include "defines.h"

class MIDIplayer {
public:
  virtual int load(char *)=0;
  virtual void init()=0;
  virtual void play()=0;
  virtual void clear()=0;
  static char stop;
  static char next;

protected:
  virtual void AllNoteOff()=0;
};

class R36_Player : public MIDIplayer{  // rcp/r36 player
public:
  R36_Player();
  virtual ~R36_Player();

  virtual int load(char *);
  virtual void init();
  virtual void play();
  virtual void clear();

protected:
  EventQueue Queue;
  EventQueue Noff_Queue; // Ρȥե٥ѤΥ塼

  ComeOn_Header Header;
  ComeOn_RythmDef Rythm;
  ComeOn_UserExclusiveDef Exclusive;
  R36_Track *Track;

  void NoteOff(Event *);
  virtual void AllNoteOff();
};

class G36_Player : public R36_Player  // g18/g36 player
{
public:
  virtual int load(char *);
};

class SMF_Player : public MIDIplayer { // SMF format0/1 player
public:
  SMF_Player();
  virtual ~SMF_Player();

  virtual int load(char *);
  virtual void init();
  virtual void play();
  virtual void clear();
  
protected:
  virtual void AllNoteOff();

  EventQueue Queue;  

  SMF_Header Header;
  SMF_Track  *Track;
};

#endif
